

//////////////////////////////////////////////////////////////////
//                                                              //
//           PLINK (c) 2005-2006 Shaun Purcell                  //
//                                                              //
// This file is distributed under the GNU General Public        //
// License, Version 2.  Please see the file COPYING for more    //
// details                                                      //
//                                                              //
//////////////////////////////////////////////////////////////////


#include <iostream>
#include <iomanip>
#include <fstream>
#include <algorithm>
#include <functional>
#include <cmath>

#include "plink.h"
#include "options.h"
#include "sets.h"


void calcGENEPIMeanVariance(vector<CSNP*> &,
			    int, int,
			    bool,
			    vector<double> &,
			    vector<vector<double> > &,
			    vector<Individual*>&,
			    vector<int> &,
			    vector<int> &);

vector<double> CCA(bool perm, 
		   vector<vector<int> > & blperm_case,
		   vector<vector<int> > & blperm_control,
		   Set & S,
		   Plink & P);

void Plink::driverSCREEPI()
{

  ///////////////////////////////
  // Gene-based epistasis
  
  printLOG("\n\n*** Warning: method under development, not for general use *** \n\n");
  

  //////////////////////////////////////////
  // Case-control samples only

  affCoding(*this);

  //////////////////////////////////////////
  // Requires that sets have been speciefied

  if (par::set_test) readSet();
  else error("Need to specify genes with --set {filename} when using --genepi\n");

    
  //////////////////
  // SET statistics

  Set S(snpset);


  //////////////////////////////////////////////
  // Prune SET (0-sized sets, MAF==0 SNPs, etc) 

  S.pruneSets(*this);
  
  int ns = snpset.size();
  if (ns < 2)
    error("Need to specify at least two fully valid sets\n");


  int ncase = 0;
  int ncontrol = 0;

  /////////////////////////////////////////////////////////
  // Prune based on VIF, separately for cases and controls

  printLOG("\nConsidering cases: ");
  setFlags(false);
  vector<Individual*>::iterator person = sample.begin();
  while ( person != sample.end() )
    {
      if ( (*person)->aff && ! (*person)->missing )
	{
	  (*person)->flag = true;
	  ncase++;
	}
      person++;
    }

  string original_outfile = par::output_file_name;
  par::output_file_name += ".case";
  S.pruneMC(*this,false,par::epi_vif);

  if (!par::epi_caseonly)
    {
      printLOG("Considering controls: ");
      setFlags(false);
      vector<Individual*>::iterator person = sample.begin();
      while ( person != sample.end() )
	{
	  if ( (!(*person)->aff) && ! (*person)->missing )
	    {
	      (*person)->flag = true;
	      ncontrol++;
	    }
	  person++;
	}
      par::output_file_name = original_outfile + ".control";
      S.pruneMC(*this,false,par::epi_vif);      
    }
  par::output_file_name = original_outfile;

  // Write finalized set
  ofstream SET1, SET2;
  string f = par::output_file_name + ".all.set.in";
  
  printLOG("Writing combined pruned-in set file to [ " + f + " ]\n");
  SET1.open(f.c_str(),ios::out);

  f = par::output_file_name + ".all.set.out";
  printLOG("Writing combined pruned-out set file to [ " + f + " ]\n");
  SET2.open(f.c_str(),ios::out);
  
  for (int s=0; s<snpset.size(); s++)
    {
      
      int nss = snpset[s].size();
      
      SET1 << setname[s] << "\n";
      SET2 << setname[s] << "\n";
      
      for (int j=0; j<nss; j++)
	{
	  if (S.cur[s][j])
	    SET1 << locus[snpset[s][j]]->name << "\n";
	  else
	    SET2 << locus[snpset[s][j]]->name << "\n";
	}
      
      SET1 << "END\n\n";
      SET2 << "END\n\n";
    }
  
  SET1.close();
  SET2.close();
  

  // Prune empty sets once more:

  S.pruneSets(*this);
  
  ns = snpset.size();
  if (ns < 2)
    error("Need to specify at least two fully valid sets\n");



  ////////////////////////////////
  // Set up permutation structure

  // Specialized (i.e. cannot use Perm class) as this 
  // requires a block-locus permutation

  // First block is fixed
  
  vector<vector<int> > blperm_case(ns);
  vector<vector<int> > blperm_control(ns);

  for (int i=0; i<ns; i++)
    {
      // A slot for each individual per locus
      for (int j=0; j<n; j++)
	if ( sample[j]->aff && ! sample[j]->missing )
	  blperm_case[i].push_back(j);
      
      for (int j=0; j<n; j++)
	if ( (!sample[j]->aff) && ! sample[j]->missing )
	  blperm_control[i].push_back(j);
      
    }




  /////////////////
  // Original data

  vector<double> original = CCA(false, 
				blperm_case, 
				blperm_control,
				S,*this);
  


  if (!par::permute) 
   return;
  
  ////////////////////////////
  // Permutation (mperm fixed)
  
  vector<double> pvalue(original.size(),0);
  
  vector<vector<int> > blorig_case = blperm_case;
  vector<vector<int> > blorig_control = blperm_control;

  for (int p=0; p<par::replicates; p++)
    {
      if (!par::silent)
	{
	  cout << "Permutation " << p+1 
	       << " of " << par::replicates 
	       << "         \r";
	  cout.flush();
	} 

      // Permute ( the second locus onwards )
      
      for (int i=1; i<ns; i++) // note: 1..ns-1
	{
	  vector<int> pi(ncase);
	  permute(pi);
	  
	  for (int j=0; j<ncase; j++)
	    blperm_case[i][j] = blorig_case[i][pi[j]];

	  vector<int> pi2(ncontrol);
	  permute(pi2);	  
	  for (int j=0; j<ncontrol; j++)
	    blperm_control[i][j] = blorig_control[i][pi2[j]];
	  
	}


      // Analyse permuted dataset

      vector<double> permuted = CCA(true, 
				    blperm_case, 
				    blperm_control,
				    S,*this);
      
      
      // Tally up times exceeded
      
      for (int j=0; j<original.size(); j++)
	{
	  
	  if ( permuted[j] > original[j] ) 
	    pvalue[j]++;
	}
      

    } // next permutation


  if (!par::silent)
    cout << "\n";



  /////////////////////////////////
  // Display results of permutation

  f = par::output_file_name+".genepi.mperm";
  printLOG("\nWriting gene-based epistasis permutation results to [ " + f + " ]\n");
  ofstream EPI(f.c_str(), ios::out);
  EPI.precision(4);

  int i=0;
  for (int s1=0; s1 < ns-1; s1++)
    for (int s2 = s1+1; s2 < ns; s2++)
      {
	EPI << setw(8) << setname[s1] << " " 
	    << setw(8) << setname[s2] << " "
	    << setw(8) << (pvalue[i]+1.0)/(double)(par::replicates+1.0) << " "
	    << "\n";
	i++;
      }
  
  EPI.close();
}




vector<double> CCA(bool perm, 
		   vector<vector<int> > & blperm_case,
		   vector<vector<int> > & blperm_control,
		   Set & S,
		   Plink & P)
  
{


  ///////////////
  // Output file

  ofstream EPI;

  if (!perm)
  {  
   string f = par::output_file_name+".genepi";
   P.printLOG("\nWriting gene-based epistasis tests to [ " + f + " ]\n");
   EPI.open(f.c_str(), ios::out);
   EPI.precision(4);
  }
  
  //////////////////////////////////
  // Canonical correlation analysis

  vector<double> res;

  int ns = P.snpset.size();

  // Consider each pair of genes

  for (int s1=0; s1 < ns-1; s1++)
    for (int s2 = s1+1; s2 < ns; s2++)
      {
	
	// make total covariance matrix in cases (controls)
	
	// partition covariance matrix
	//
	//   S_11  S_21
	//   S_12  S_22
	
	int n1=0, n2=0;
	
	vector<vector<double> > sigma_case(0);
	vector<vector<double> > sigma_control(0);
	vector<double> mean(0);
	vector<CSNP*> pSNP(0);

	/////////////////////////////
	// List of SNPs for both loci

	for (int l=0; l<P.snpset[s1].size(); l++)
	  if ( S.cur[s1][l] )
	    {
	      pSNP.push_back( P.SNP[ P.snpset[s1][l] ] );
	      n1++;
	    }
	for (int l=0; l<P.snpset[s2].size(); l++)
	  if ( S.cur[s2][l] )
	    {
	      pSNP.push_back( P.SNP[ P.snpset[s2][l] ] );
	      n2++;
	    }

	int n12 = n1 + n2;

	///////////////////////////////////
	// Construct case covariance matrix

	P.setFlags(false);
	vector<Individual*>::iterator person = P.sample.begin();
	while ( person != P.sample.end() )
	  {
	    if ( (*person)->aff && !(*person)->missing )
	      (*person)->flag = true;
	    person++;
	  }

	calcGENEPIMeanVariance(pSNP, 
			       n1,n2,
			       false,
			       mean, 
			       sigma_case, 
			       P.sample , 
			       blperm_case[s1], 
			       blperm_case[s2] );
	
	if (!par::epi_caseonly)
	  {
	    P.setFlags(false);
	    vector<Individual*>::iterator person = P.sample.begin();
	    while ( person != P.sample.end() )
	      {
		if ( (!(*person)->aff) && !(*person)->missing )
		  (*person)->flag = true;
		person++;
	      }
	    calcGENEPIMeanVariance(pSNP, 
				   n1,n2,
				   false,
				   mean, 
				   sigma_control, 
				   P.sample, 
				   blperm_control[s1],
				   blperm_control[s2] );
	    
	  }


	///////////////////////////
	// Cases

	vector<vector<double> > I11_case;
	vector<vector<double> > I12_case;
	vector<vector<double> > I21_case;
	vector<vector<double> > I22_case;
	
	sizeMatrix( I11_case, n1, n1);
	sizeMatrix( I12_case, n1, n2);
	sizeMatrix( I21_case, n2, n1);
	sizeMatrix( I22_case, n2, n2);

	for (int i=0; i<n1; i++)
	  for (int j=0; j<n1; j++)
	    I11_case[i][j] = sigma_case[i][j];

	for (int i=0; i<n1; i++)
	  for (int j=0; j<n2; j++)
	    I12_case[i][j] = sigma_case[i][n1+j];

	for (int i=0; i<n2; i++)
	  for (int j=0; j<n1; j++)
	    I21_case[i][j] = sigma_case[n1+i][j];

	for (int i=0; i<n2; i++)
	  for (int j=0; j<n2; j++)
	    I22_case[i][j] = sigma_case[n1+i][n1+j];


	// calculate canonical correlations
	// M1 <- solve(msqrt(S11)) %*% S12 %*% solve(S22) %*% S21 %*% solve(msqrt(S11))  
	// E1 <- eigen(M1)
	// cat("canonical correlations",sqrt(E1$val),"\n")	
	
	I11_case = msqrt(I11_case);
	I11_case = svd_inverse(I11_case);
	I22_case = svd_inverse(I22_case);
	
	vector<vector<double> > tmp;
	vector<vector<double> > tmp2;

	multMatrix(I11_case, I12_case, tmp);
	multMatrix(tmp, I22_case, tmp2);
	multMatrix(tmp2, I21_case, tmp);
	multMatrix(tmp, I11_case, tmp2);

	vector<double> eig_case = eigenvalues(tmp2);
	
	int ne = n1 < n2 ? n1 : n2;
	
	sort(eig_case.begin(),eig_case.end(),greater<double>());
	
	for (int i=0; i<ne; i++)
	  {
	    if ( eig_case[i] < 0 ) 
	      eig_case[i] = 0;
	    else if ( eig_case[i] >= 1 )
	      eig_case[i] = 0.999;
	  }


	// Case-only statistics
	// Test all, 2-N, 3-N, etc
	vector<double>SN_case(ne,0);
	for (int i=0; i<ne; i++)
	  {
	    double prod = 1;
	    for (int j=i; j<ne; j++)
	      prod *= 1 - eig_case[j];
	    SN_case[i] = -log(prod);
	  }
	     
	
	////////////////////////////
	// Controls
	
	vector<double> SN_control;
	
	if (!par::epi_caseonly)
	  {
	    
	    vector<vector<double> > I11_control;
	    vector<vector<double> > I12_control;
	    vector<vector<double> > I21_control;
	    vector<vector<double> > I22_control;
	
	    sizeMatrix( I11_control, n1, n1);
	    sizeMatrix( I12_control, n1, n2);
	    sizeMatrix( I21_control, n2, n1);
	    sizeMatrix( I22_control, n2, n2);
	    
	    for (int i=0; i<n1; i++)
	      for (int j=0; j<n1; j++)
		I11_control[i][j] = sigma_control[i][j];
	    
	    for (int i=0; i<n1; i++)
	      for (int j=0; j<n2; j++)
		I12_control[i][j] = sigma_control[i][n1+j];
	    
	    for (int i=0; i<n2; i++)
	      for (int j=0; j<n1; j++)
		I21_control[i][j] = sigma_control[n1+i][j];
	    
	    for (int i=0; i<n2; i++)
	      for (int j=0; j<n2; j++)
		I22_control[i][j] = sigma_control[n1+i][n1+j];
	    

	    I11_control = msqrt(I11_control);
	    I11_control = svd_inverse(I11_control);
	    I22_control = svd_inverse(I22_control);
	    
	    vector<vector<double> > tmp;
	    vector<vector<double> > tmp2;
	    
	    multMatrix(I11_control, I12_control, tmp);
	    multMatrix(tmp, I22_control, tmp2);
	    multMatrix(tmp2, I21_control, tmp);
	    multMatrix(tmp, I11_control, tmp2);
	    
	    vector<double> eig_control = eigenvalues(tmp2);
	    
	    int ne = n1 < n2 ? n1 : n2;
	    
	    sort(eig_control.begin(),eig_control.end(),std::greater<double>());
		
	    for (int i=0; i<ne; i++)
	      {
		if ( eig_control[i] < 0 ) 
		  eig_control[i] = 0;
		else if ( eig_control[i] >= 1 )
		  eig_control[i] = 0.999;
	      }
	    
	    // Control-only statistics
	    SN_control.resize(ne,0);
	    for (int i=0; i<ne; i++)
	      {
		double prod = 1;
		for (int j=i; j<ne; j++)
		  prod *= 1 - eig_control[j];
		SN_control[i] = -log(prod);
	      }
	
	     
	  }
      
	
	// Statistics
	
	vector<double> SN(ne);
	if (!par::epi_caseonly)
	  {	
	    for (int i=0; i<ne; i++)
	      SN[i] = SN_case[i] - SN_control[i];
	  }
	else
	  SN = SN_case;
	
// 	cout << "chisq = "  
// 	     << setw(10) << P.setname[s1] << " " 
// 	     << setw(10) << P.setname[s2] << " " 
// 	     << n1 << " " 
// 	     << n2 << "\t"
// 	     << SN[0] << "\t" 
// 	     << SN[ne-1] << "\t"
// 	     << eig_case[0] << "\n";
	
	// Store for permutation: either largest test statistic
	res.push_back(SN[0]);
	
	if (!perm)
	  {
	    for (int i=0; i<ne; i++)
	      {
		
		bool maxchisq = false;	      
		double chisq = (blperm_case[s1].size()-1-( double(n1+n2+1)/2.0 )) * SN_case[i];
		if (chisq<0) chisq=0;
		else if (chisq>500) { chisq=500; maxchisq = true; } 
		
		int df = (n1-i)*(n2-i);
		double pvalue = chiprobP( chisq , df );
		
		EPI << setw(8) << P.setname[s1] << " " 
		    << setw(4) << n1 << " "
		    << setw(8) << P.setname[s2] << " "
		    << setw(4) << n2 << " "
		    << setw(4) << i << " "
		    << setw(8) << sqrt(eig_case[i]) << " ";
		if (!maxchisq) EPI << setw(8) << chisq << " ";
		else EPI << setw(8) << "500+" << " ";
		EPI << setw(8) << df << " "	    
		    << setw(8) << pvalue << " "
		    << "\n";
		
	      }
	    EPI <<"\n";
	  }
      }
  

  if (!perm) EPI.close();
  
  return res;
  cout << "\n";
}



void calcGENEPIMeanVariance(vector<CSNP*> & pSNP,
			    int n1,
			    int n2,
			    bool perm, 
			    vector<double> & mean,
			    vector<vector<double> > & variance,
			    vector<Individual*> & sample,
			    vector<int> & gp1,
			    vector<int> & gp2 )
  
{

  // Calculate mean and variance for n1+n2 x n1+n2 matrix

  // Individual order in n1 , n2 deteremined by g1, g2
  // (i.e. block-based permutation)
  
  // Under permutations, mean and variances won't change
  // Store means only for now

  int n = gp1.size();
  
  int nss = pSNP.size();
  
  // Original calculation?
  if (!perm)
    mean.resize(nss,0);
  
  vector<int> cnt(nss,0);
  variance.resize(nss);
  
  for (int j=0; j<nss; j++)
    variance[j].resize(nss,0);
  
      
  /////////
  // Means

  /////////////////////////////////
  // Consider each SNP in this set
  
  for (int j=0; j<nss; j++)
    {
      
      CSNP * ps = pSNP[j];
      
      ///////////////////////////
      // Iterate over individuals
      
      for (int i=0; i<n; i++)
	{	      
	  // Only need to look at one perm set
	  bool a1 = ps->one[gp1[i]];
	  bool a2 = ps->two[gp2[i]];
	  
	  if ( a1 )
	    {		      
	      if ( a2 ) // 11 homozygote
		{
		  mean[j]++;
		  cnt[j]++;
		}
	    }
	  else 
	    {
	      cnt[j]++;
	      if ( ! a2  ) // 00 homozygote
		mean[j]--;
	      
	    }
	  
	} // Next individual	  
      
    } // Next SNP in set
  
  for (int j=0; j<nss; j++)
    mean[j] /= (double)cnt[j];
  
  
  
  /////////////////////////////////////
  // Iterate over pairs of SNPs in SETs
  
  // First SNP 
  for (int j1=0; j1<nss; j1++)
    {
      CSNP * ps1 = pSNP[j1];

      // Second SNP
      for (int j2=0; j2<nss; j2++)
	{
	  CSNP * ps2 = pSNP[j2];

	  // Iterate over individuals
	  
	  for (int i=0; i<n; i++)
	    {
	      
	      bool a1, a2;
	      if (j1<n1)
		{
		  a1 = ps1->one[gp1[i]];
		  a2 = ps1->two[gp1[i]];
		}
	      else
		{
		  a1 = ps1->one[gp2[i]];
		  a2 = ps1->two[gp2[i]];
		}
	      
	      bool b1, b2;
	      if (j1<n1)
		{
		  b1 = ps2->one[gp1[i]];
		  b2 = ps2->two[gp1[i]];
		}
	      else
		{
		  b1 = ps2->one[gp2[i]];
		  b2 = ps2->two[gp2[i]];
		}
	      
	      
	      // Mean substitution
	      double v1=mean[j1], v2=mean[j2];
	      
	      // First SNP
	      if ( a1 )
		{		      
		  if ( a2 )    // 11 homozygote
		    {
		      v1 = 1;
		    }
		}
	      else 
		{
		  if ( ! a2  ) // 00 homozygote
		    {
		      v1 = -1;
		    }
		  else
		    v1 = 0;       // 01 heterozygote
		}
	      
	      
	      // Second SNP
	      if ( b1 )
		{		      
		  if ( b2 )    // 11 homozygote
		    {
		      v2 = 1;
		    }
		}
	      else 
		{
		  if ( ! b2  ) // 00 homozygote
		    {
		      v2 = -1;
		    }
		  else
		    v2 = 0;       // 01 heterozygote
		}
	      
	      
	      // Contribution to covariance term
	      variance[j1][j2] += ( v1 - mean[j1] ) * ( v2 - mean[j2] );
	      
		} // Next individual
	    } // Second SNP
	} // First SNP
        
      // Make symmetric covariance matrix
      for (int i=0; i<nss; i++)
	for (int j=i; j<nss; j++)
	  {
	    variance[i][j] /= (double)(n);
	    variance[j][i] = variance[i][j];
	  }
      
      return;
      
}
  
