


//////////////////////////////////////////////////////////////////
//                                                              //
//           PLINK (c) 2005-2006 Shaun Purcell                  //
//                                                              //
// This file is distributed under the GNU General Public        //
// License, Version 2.  Please see the file COPYING for more    //
// details                                                      //
//                                                              //
//////////////////////////////////////////////////////////////////


#include <iostream>
#include <iomanip>
#include <fstream> 
#include <map>
#include <vector>
#include <set>
#include <cmath>

#include "plink.h"
#include "options.h"
#include "crandom.h"
#include "sets.h"
#include "perm.h"

////////////////////////////////////////////////////////
// Helper function: add individual to family, w/ checks.

void addParent(Family * f, Individual * person)
{
  // A real person? 
  if ( person )
    {
      // as father
      if ( person->sex )
	{
	  if ( f->pat == NULL ) { 
	    f->pat = person; 

	    // Do not overwrite (*see note below)
	    if (!person->family)
	      person->family = f; 
	  } 
	  else error("Problem pedigree structure: two fathers found : family "
		     +person->fid);
	  
	}
      else if (person->sexcode=="2")
	{   
	  // as mother
	  if ( f->mat == NULL ) { 
	    f->mat = person; 

	    // Do not overwrite (*see note below)
	    if (!person->family)
	      person->family = f; 
	  }
	  else error("Problem pedigree structure: two mothers found: family "
		     +person->fid);
	}
      else
	{
	  error("Problem with ambiguous parental sex codes for family "
		+person->fid);
	}
    }
  else
    {
      // Otherwise, add as a dummy individual
      if ( !f->pat )
	f->pat = person;
      else if ( ! f->mat )
	f->mat = person;
      else error("Internal error: allocated too many parents...\n");
    }
      
}

void addPerson(Family * f, Individual * person)
{
  
  // Add as child if does not already exist
  
  for (int c=0; c<f->kid.size(); c++)
    if ( f->kid[c]->iid == person->iid ) 
      error("Problem with family "
	    +f->kid[c]->fid+" child "+f->kid[c]->iid
	    +": offspring already exists\n");
  f->kid.push_back(person);
  // Always set (*see note)
  person->family = f;
}


// Note on priority of setting person->family pointer:
// This will be preferentially set to the family in which 
// the individual is the offspring (i.e. an individual can 
// only appear in one nuclear family as an offspring, but 
// multiple as a founder). Therefore, if nothing has been set 
// (singleton) set the family pointer for founder (i.e. so as 
// not to overwrite offspring family pointer) but otherwise
// always overwrite if setting the offspring family.

// In qfam.cpp then all family[] are used to construct B scores
// (i.e. an individual's genotypes might be used in several families)
// but the person->family is used to enter the person into the actual
// analysis (i.e. so the individual will always appear as offspring,
// or sibship (even is S=1)


void Plink::parseTrios()
{  

  /////////////////////////////////////
  // General check for unique FID, IIDs
  // and that no IID == 0 

  set<string> fid_iid;
  vector<Individual*>::iterator person = sample.begin();
  while ( person != sample.end() )
    {
      if ( (*person)->iid == "0" )
	error("Family "+(*person)->fid+" has person with reserved 0 ID\n");
      
      string s = (*person)->fid+"_"+(*person)->iid;
      if ( fid_iid.find(s) != fid_iid.end() )
	error("Duplicate individual in pedigree: "
	      +(*person)->fid+" "
	      +(*person)->iid+"\n");
      else
	fid_iid.insert(s);
      person++;
    }
  


  /////////////////////////////////////
  // First consider all nonfounders 
  // with 2 parents

  map<string,Family*> fam;
  map<string,Family*>::iterator f;
  set<Individual*> infamily;
  
  person = sample.begin();
  
  while ( person != sample.end() )
    {
      
      // For non-founders
      if ( ! (*person)->founder ) 
	{
	  
	  string fpm = (*person)->fid+
	    "_"+(*person)->pat+
	    "_"+(*person)->mat;
	  
	  f = fam.find(fpm);
	  
	  // Have we already come across this parental pairing?
	  if ( f != fam.end() ) 
	    {
	      addPerson(f->second,*person);
	      infamily.insert(*person);
	    }
	  else 
	    {
	      Family * nfam = new Family;
	      
	      // Add person to new family
	      addPerson(nfam,*person);

	      infamily.insert(*person);

	      // And the parental pairing
	      
	      addParent(nfam,(*person)->pp);
	      addParent(nfam,(*person)->pm);		  
	      
	      infamily.insert( (*person)->pp );
	      infamily.insert( (*person)->pm );

	      if ( (*person)->pp != NULL && 
		   (*person)->pm != NULL )
		nfam->parents = true;
	      else
		nfam->sibship = true;
	      
	      // And add it to the list
	      fam.insert(make_pair(fpm,nfam));	  
	      
	    }
	}
      
      
      // Consider next individual
      person++;
    }
  
  // Rescan list for singleton founders
  person = sample.begin();
  while ( person != sample.end() )
    {
      if ( infamily.find( *person ) == infamily.end() )
	{
	  Family * nfam = new Family;
	  
	  // Have we already seen a singleton founder in this family?
	  map<string,Family*>::iterator myf = fam.find( (*person)->fid+"_0_0" );
	  if ( myf != fam.end() )
	    nfam = myf->second;
	      
	  addPerson(nfam,*person);
	  nfam->singleton = true;
	  fam.insert(make_pair((*person)->fid+"_0_0",nfam));	  
	}
      person++;
    }  
  

  // Counts
  int disc_parent_cnt = 0;
  int with_parents_ind_cnt = 0;
  int with_parents_fam_cnt = 0;
  int aff_with_parents_trio_cnt = 0;
  int without_parents_ind_cnt = 0;
  int without_parents_fam_cnt = 0;
  int singleton_cnt = 0;

  // Assign family type flags
  
  for ( f = fam.begin() ; f != fam.end() ; f++)
    {
      
      Family * mf = f->second;
      
      // TDT = 2 parents + atleast 1 affected offspring
      
      if ( mf->kid.size()>=0) 
	{
	  
	  if ( mf->parents )
	    {
	      with_parents_fam_cnt++;
	      for (int k=0; k<mf->kid.size(); k++)
		{
		  with_parents_ind_cnt++;
		  if (mf->kid[k]->aff)
		    {
		      aff_with_parents_trio_cnt++;
		      mf->TDT = true;
		    }
		}
	      
	    }
	  else
	    {
	      without_parents_fam_cnt++;
	      for (int k=0; k<mf->kid.size(); k++)
		without_parents_ind_cnt++;
	    }
	  
	  // Set flag for phenotypically discordant parents
	  
	  mf->discordant_parents = false;
	  if ( mf->parents )
	    if ( mf->pat->phenotype != mf->mat->phenotype &&
		 (!mf->pat->missing)  && 
		 (!mf->mat->missing)  )
	      {
		mf->discordant_parents = true;
		disc_parent_cnt++;
	      }
	} 
      
      if (mf->singleton) 
	singleton_cnt++;

      family.push_back(mf);
      
    }
  
  

  


  
  // Report counts
  
  printLOG(int2str(family.size())+" nuclear families, ");
  printLOG(int2str(singleton_cnt)+" founder singletons found\n");
  printLOG(int2str(with_parents_ind_cnt)+" individuals with 2 parents in "
	   +int2str(with_parents_fam_cnt)+" nuclear families\n");
  printLOG(int2str(without_parents_ind_cnt-singleton_cnt)+" individuals without 2 parents in "
	   +int2str(without_parents_fam_cnt-singleton_cnt)+" nuclear families\n");

  if (par::bt)
    printLOG(int2str(aff_with_parents_trio_cnt)+" affected offspring trios\n");
	      
  printLOG(int2str(disc_parent_cnt)+" phenotypically discordant parent pairs found\n");
  
  if (disc_parent_cnt>=10) par::discordant_parents = true;

  
  if ( par::dumpped ) 
    {
      string str = par::output_file_name + ".pdump";
      printLOG("Dumping pedigree information to [ " + str + " ]\n");
      ofstream PD(str.c_str(),ios::out);

      vector<Family*>::iterator f = family.begin();
     
      while ( f != family.end() )
	{

	  if ( (*f)->singleton )
	    {
	      PD << "SINGLETON(S)\t" 
		 << (*f)->kid[0]->fid << " : ";
		for (int k=0; k < (*f)->kid.size() ;k++)
		  PD << (*f)->kid[k]->iid << " ";
	      PD << "\n";
	    }
	  else if ( (*f)->sibship )
	    {
	      PD << "SIBSHIP  \t" << (*f)->kid[0]->fid << " : ";
	      for ( int k=0; k<(*f)->kid.size(); k++)
		PD << (*f)->kid[k]->iid << " ";
	      PD << "\n";
	    }
	  else if ( (*f)->parents )
	    {
	      PD << "W/ PARENTS\t" << (*f)->pat->fid << " : ";
	      PD << (*f)->pat->iid << " x " << (*f)->mat->iid << " -> ";
	      for ( int k=0; k<(*f)->kid.size(); k++)
		PD << (*f)->kid[k]->iid << " ";
	      PD << "\n";	     
	    }
	  else
	    PD << "UNDEFINED\t" 
		 << (*f)->pat->fid << " "
		 << (*f)->pat->iid << "\n";
	  
	  // Next family
	  f++;
	}
      
      PD << "\n\n";
      
      PD << "Listing by individual: columns are (0/1 for true false) \n"
	 << "  FID\tFamily ID\n"
	 << "  IID\tIndividual ID\n"
	 << "  Phenotype\n"
	 << "  Parents?\n"
	 << "  Singleton?\n"
	 << "  Sibship?\n"
	 << "  Discordant parents?\n"
	 << "  TDT?\n\n";
      
      // Now list by individual 
      
      vector<Individual*>::iterator person = sample.begin();

      while ( person != sample.end() ) 
	{
	  
	  PD << (*person)->fid << "\t"
	     << (*person)->iid << "\t"
	     << (*person)->phenotype << "\t"
	     << (*person)->family->parents << " "
	     << (*person)->family->singleton << " "
	     << (*person)->family->sibship << " " 
	     << (*person)->family->discordant_parents << " " 
	     << (*person)->family->TDT << "\n";
	    	    
	  person++;
	}

      PD.close();
    }
  

}



void Plink::checkMendel()
{


  //////////////////////////////////
  // Individual-major mode analysis
  
  if (par::SNP_major) SNP2Ind();

  ofstream MEN;
  ofstream MENL;
  ofstream MENI;
  ofstream MENF;

  if (par::MENDEL_test)
    printLOG("Filtering SNPs/families for Mendel Error rates above "+
	     dbl2str(par::MENDEL_snp)+", "+dbl2str(par::MENDEL_ind)+"\n");


  if (par::MENDEL_report)
    {
      string f = par::output_file_name+".mendel";
      string fi = par::output_file_name+".imendel";
      string ff = par::output_file_name+".fmendel";
      string fl = par::output_file_name+".lmendel";
      
      printLOG("Writing all Mendel errors to [ " + f +" ]\n");
      printLOG("Writing per-offspring Mendel summary to [ " + fi + " ]\n");
      printLOG("Writing per-family Mendel summary to [ " + ff + " ]\n");
      printLOG("Writing per-locus Mendel summary to [ " + fl + " ]\n");
      
      MEN.open(f.c_str(),ios::out);
      MENL.open(fl.c_str(),ios::out);
      MENI.open(fi.c_str(),ios::out);
      MENF.open(ff.c_str(),ios::out);
      
      MEN << setw(par::pp_maxfid) << "FID" << " " 
	  << setw(par::pp_maxiid) << "KID" << " " 
	  << setw(4) << "CHR" << " " 
	  << setw(par::pp_maxsnp) << "SNP" << " " 
	  << setw(6) << "CODE"   
	  << setw(22) << "ERROR"  
	  << "\n";
      
      MENL << setw(4) << "CHR" << " " 
	   << setw(par::pp_maxsnp) << "SNP" << " " 
	   << setw(4) << "N\n";
      
      MENI << setw(par::pp_maxfid) << "FID"  << " " 
	   << setw(par::pp_maxiid) << "IID" << " " 
	   << setw(4) << "N\n";
      
      MENF << setw(par::pp_maxfid) << "FID" << " " 
	   << setw(par::pp_maxiid) << "PAT" << " "
	   << setw(par::pp_maxiid) << "MAT" << " "
	   << setw(6) << "CHLD" << " " 
	   << setw(4) << "N" << "\n";
    }


  // Flag to remove SNP if need be
  vector<bool> mendel_locus(nl_all,false);
  // Counts per family
  vector<int> mendel_family(family.size(),0);
  vector<int> mendel_pat(family.size(),0);
  vector<int> mendel_mat(family.size(),0);
  vector<vector<int> > mendel_indiv(family.size());
  for (int f=0; f<family.size(); f++) 
    mendel_indiv[f].resize(family[f]->kid.size());
  
  // Count number of trios
  int n_trios = 0; 
  for (int f=0; f<family.size(); f++) 
    n_trios += family[f]->kid.size();
  
  int total_m = 0; // total number of Mendel errors
  int l_removed = 0; // # of SNPs removed due to high ME rate

  // Test each locus
  for (int l=0; l<nl_all; l++)
    {
      
      // Skip haploid markers
      if (par::chr_haploid[locus[l]->chr]) continue;
      
      // Flag for X markers
      bool X = par::chr_sex[locus[l]->chr];
      
      int m=0; // number of Mendel errors per locus
      
      for (int f=0; f<family.size(); f++) 
	{
	  
	  // Have we observed parents?
	  if ( ! family[f]->parents ) continue;
	  
	  Individual * pat = family[f]->pat;
	  Individual * mat = family[f]->mat;
	  vector<Individual *> kid = family[f]->kid;
	  
	  mendel_indiv[f].resize(kid.size());

	  for (int c=0; c<kid.size(); c++)
	    {
	      
	      // If child genotype is missing, skip
	      if ( kid[c]->one[l] && !kid[c]->two[l] ) continue;

	      int mendel_type = 0;
	      
	      // For autosomal markers
	      if ( (!X) || (!kid[c]->sex) )
		{
		  
		  if ( (!kid[c]->one[l]) && kid[c]->two[l] )
		    {
		      
		      // KID = 01
		      // 00x00 -> 01  (m1)
		      // 11x11 -> 01  (m2)
		      
		      
		      if (  ( (!pat->one[l]) && (!pat->two[l]) ) && 
			    ( (!mat->one[l]) && (!mat->two[l]) ) )
			{
			  kid[c]->one[l] = true;
			  kid[c]->two[l] = false;
			  m++;
			  mendel_type = 1;
			}
		      else if (  ( pat->one[l] && pat->two[l] ) && 
				 ( mat->one[l] && mat->two[l] ) )
			{
			  kid[c]->one[l] = true;
			  kid[c]->two[l] = false;
			  m++;
			  mendel_type = 2;
			}
		    }
		  else if ( (!kid[c]->one[l]) && (!kid[c]->two[l]) )
		    {
		      // KID = 00
		      // 00x11 -> 00 (m3) P11->00
		      // 01x11 -> 00 (m3) 
		      // ??x11 -> 00 (m3)
		      
		      // 11x00 -> 00 (m4) M11->00
		      // 11x01 -> 00 (m4) 
		      // 11x?? -> 00 (m4) 
		      
		      // 11x11 -> 00 (m5) P11+M11->00 
		      
		      // Hom parent --> opposite hom child
		      
		      // rule = at least one '11' parent
		      
		      if ( ( pat->one[l] && pat->two[l] ) || 
			   ( mat->one[l] && mat->two[l] ) )
			{
			  kid[c]->one[l] = true;
			  kid[c]->two[l] = false;
			  m++;
			  
			  if ( pat->one[l] && pat->two[l] && 
			       mat->one[l] && mat->two[l] )
			    mendel_type = 5;
			  else if ( pat->one[l] && pat->two[l] )
			    mendel_type = 3;
			  else
			    mendel_type = 4;		      
			}
		    }	
		  else 
		    {
		      
		      // KID = 11
		      
		      // 00x01 -> 11 (m6)
		      // 00x11 -> 11
		      // 00x?? -> 11
		      
		      // 01x00 -> 11 (m7)
		      // 11x00 -> 11
		      // ??x00 -> 11
		      
		      // 00x00 -> 11 (m8) P00+M00->11		 
		      
		      // rule = at least one '00' parent
		      
		      if ( ( (!pat->one[l]) && (!pat->two[l]) ) || 
			   ( (!mat->one[l]) && (!mat->two[l]) ) )
			{
			  kid[c]->one[l] = true;
			  kid[c]->two[l] = false;
			  m++;
			  
			  if ( (!pat->one[l]) && (!pat->two[l]) && 
			       (!mat->one[l]) && (!mat->two[l]) )
			    mendel_type = 8;
			  else if ( (!pat->one[l]) && (!pat->two[l]) )
			    mendel_type = 6;
			  else
			    mendel_type = 7;		      
			  
			}
		      		      
		    }
		  
		}
	      else
		{
		  if ( kid[c]->one[l] && kid[c]->two[l] && 
		       (!mat->one[l]) && (!mat->two[l]) )
		    {
		      m++;
		      mendel_type = 9;
		      kid[c]->one[l] = true;
		      kid[c]->two[l] = false;

		    }

		  if ( (!kid[c]->one[l]) && (!kid[c]->two[l]) && 
		       mat->one[l] && mat->two[l] )
		    {
		      m++;
		      mendel_type = 10;
		      kid[c]->one[l] = true;
		      kid[c]->two[l] = false;		      
		    }
		      
		}
		    
	      // Individual counts
	      
	      // m1   00x00  ->  01   K / P+M  
	      // m2   11x11  ->  01   K / P+M  

	      // m3   11x**  ->  00   K / P    
	      // m4   **x11  ->  00   K / M    
	      // m5   11x11  ->  00   K / P+M  

	      // m6   00x**  ->  11   K / P    
	      // m7   **x00  ->  11   K / M    
	      // m8   00x00  ->  11   K / P+M  

	      // X marker errors for male offspring
	      // m9   **x00 -> 11     K / M    
	      // m10  **x11 -> 00     K / M
	      
	      
	      if (par::MENDEL_report && mendel_type>0)
		{
		  MEN << setw(par::pp_maxfid) << family[f]->pat->fid  << " " 
		      << setw(par::pp_maxiid) << kid[c]->iid << " " 
		      << setw(4)  << locus[l]->chr  << " " 
		      << setw(par::pp_maxsnp) << locus[l]->name  << " " ;
		  
		  if (mendel_type==1) 
		    {
		      MEN << setw(6) << "1"<< setw(22);
		      string s = locus[l]->allele1 + "/" + locus[l]->allele1 + " x " 
			+ locus[l]->allele1 + "/" + locus[l]->allele1 + " -> " 
			+ locus[l]->allele1 + "/" + locus[l]->allele2;
		      MEN << s << "\n";
		    }
		  else if (mendel_type==2)
		    {
		      MEN << setw(6) << "2"<< setw(22);
		      string s = locus[l]->allele2 + "/" + locus[l]->allele2 + " x "
			+ locus[l]->allele2 + "/" + locus[l]->allele2 + " -> "
			+ locus[l]->allele1 + "/" + locus[l]->allele2;
		      MEN << s << "\n";
		    }
		  else if (mendel_type==3)
		    {
		      MEN << setw(6) << "3"<< setw(22);
		      string s = locus[l]->allele2 + "/" + locus[l]->allele2 + " x "
			+ "*/*" + " -> "
			+ locus[l]->allele1 + "/" + locus[l]->allele1;
		      MEN << s << "\n";
		    } 
		  else if (mendel_type==4)
		    {	
		      MEN << setw(6) << "4"<< setw(22);
		      string s = string("*/*") + string(" x ") 
			+ locus[l]->allele2 + "/" + locus[l]->allele2 + " -> "
			+ locus[l]->allele1 + "/" + locus[l]->allele1;
		      MEN << s << "\n";
		    }	
		  else if (mendel_type==5)
			{	
			  MEN << setw(6) << "5"<< setw(22);
			  string s= locus[l]->allele2 + "/" + locus[l]->allele2 + " x "
			    + locus[l]->allele2 + "/" + locus[l]->allele2 + " -> "
			    + locus[l]->allele1 + "/" + locus[l]->allele1;
			  MEN << s << "\n";
			}	
		  else if (mendel_type==6)
		    {
		      MEN << setw(6) << "6"<< setw(22);
		      string s = locus[l]->allele1 + "/" + locus[l]->allele1 + " x "
 			+ "*/*" + " -> "
			+ locus[l]->allele2 + "/" + locus[l]->allele2;
		      MEN << s << "\n";
		    }
		  else if (mendel_type==7)
		    {
		      MEN << setw(6) << "7"<< setw(22);
		      string s = string("*/*") + string(" x ") + locus[l]->allele1 + "/" + locus[l]->allele1 + " -> "
			+ locus[l]->allele2 + "/" + locus[l]->allele2;
			MEN << s << "\n";
		    }
		  else if (mendel_type==8)
		    {
		      MEN << setw(6) << "8" << setw(22); 
		      string s = locus[l]->allele1 + "/" + locus[l]->allele1 + " x "
			+ locus[l]->allele1 + "/" + locus[l]->allele1 + " -> "
			+ locus[l]->allele2 + "/" + locus[l]->allele2;
			MEN << s << "\n";
		    }
		  else if (mendel_type==9)
		    {
		      MEN << setw(6) << "9" << setw(22); 
		      string s = string("*/*") + " x "
			+ locus[l]->allele1 + "/" + locus[l]->allele1 + " -> "
			+ locus[l]->allele2 + "/" + locus[l]->allele2;
		      MEN << s << "\n";
		    }
		  else if (mendel_type==10)
		    {
		      MEN << setw(6) << "10" << setw(22); 
		      string s = string("*/*") + " x "
			+ locus[l]->allele2 + "/" + locus[l]->allele2 + " -> "
			+ locus[l]->allele1 + "/" + locus[l]->allele1;
		      MEN << s << "\n";
		    }

		}
	      
	      
	      // Family count
	      if (mendel_type>0) 
		{
		  // We may wish to add other weighting schemes here
		  
		  if ( mendel_type==1 || mendel_type==2 ) 
		    { 
		      mendel_indiv[f][c]++; 
		      mendel_pat[f]++; 
		      mendel_mat[f]++; 
		    }
		  else if ( mendel_type==5 || mendel_type==8 ) 
		    mendel_indiv[f][c]++;
		  else if ( mendel_type==3 || mendel_type==6 ) 
		    { 
		      mendel_indiv[f][c]++; 
		      mendel_pat[f]++; 
		    }
		  else if ( mendel_type==4 || mendel_type==7 ) 
		    { 
		      mendel_indiv[f][c]++; 
		      mendel_mat[f]++; 
		    }	      
		  else if ( mendel_type==9 || mendel_type==10 ) 
		    { 
		      mendel_indiv[f][c]++; 
		      mendel_mat[f]++; 
		    }
		}

	      // Family count
	      if (mendel_type>0) 
		mendel_family[f]++;
	      
	    }

	}


      // Report per-SNP Mendel count 
      if (par::MENDEL_report)
	MENL << setw(4)  << locus[l]->chr << " "
	     << setw(par::pp_maxsnp) << locus[l]->name << " "
	     << setw(4)  << m << "\n";

      // Is this a bad SNP? 
      if ( (double)m/(double)n_trios > par::MENDEL_snp)
	{
	  l_removed++;
	  mendel_locus[l] = true;
	}

      // Keep track of total number of errors
      total_m += m;
      
    }


  printLOG(int2str(total_m)+" Mendel errors detected in total\n");

  if (par::MENDEL_report)
    {  
      for (int f=0; f<family.size(); f++) 
	{
	  if ( family[f]->parents )
	    MENF << setw(par::pp_maxfid) << family[f]->pat->fid << " "
		 << setw(par::pp_maxiid) << family[f]->pat->iid << " "
		 << setw(par::pp_maxiid) << family[f]->mat->iid << " "
		 << setw(6) << family[f]->kid.size() << " "
		 << setw(4) << mendel_family[f]  
		 << "\n";
	}
      
      for (int f=0; f<family.size(); f++) 
	{
	  if ( family[f]->parents )
	    {
	      vector<Individual *> kid = family[f]->kid;
	      MENI << setw(par::pp_maxfid) << family[f]->pat->fid << " "
		   << setw(par::pp_maxiid) << family[f]->pat->iid << " "
		   << setw(4)  << mendel_pat[f] 
		   << "\n";
	      MENI << setw(par::pp_maxfid) << family[f]->mat->fid << " "
		   << setw(par::pp_maxiid) << family[f]->mat->iid << " "
		   << setw(4)  << mendel_mat[f]  << " "
		   << "\n";
	      for (int c=0; c<kid.size(); c++)
		MENI << setw(par::pp_maxfid) << family[f]->pat->fid << " "
		     << setw(par::pp_maxiid) << kid[c]->iid << " "
		     << setw(4)  << mendel_indiv[f][c] 
		     << "\n";
	    }      
	}

      MEN.close();
      MENL.close();
      MENI.close();
      MENF.close();

      shutdown();
    }



  // Using Mendel error rates to automatically remove SNPs 
  // and families
  
  if (par::MENDEL_test)
    {
      
      ////////////////////////////////////////
      // Remove selected loci from locus list, 
      // by copying rest to a new list
      // People/genotypes first, then locus/map info

      bool any_mendel = false;
      for (int l=0; l<nl_all; l++)
	if ( mendel_locus[l] ) any_mendel = true;
      
      int orig_nl = nl_all; // copy for comparison below

      if (any_mendel)
	int tmp = deleteSNPs(mendel_locus);


      map<Individual*,int> badfam;
      int f_removed = 0;
      for (int f=0;f<family.size();f++)
	{
	  if ((double)mendel_family[f]/(double)orig_nl > par::MENDEL_ind) 
	    {
	      f_removed++;
	      badfam.insert(make_pair(family[f]->pat,0));
	      badfam.insert(make_pair(family[f]->mat,0));
	      for (int c=0; c<family[f]->kid.size(); c++)
		badfam.insert(make_pair(family[f]->kid[c],0));
	    }
	}
      
      // Remove individuals as appropriate
      int n_removed = 0;
      vector<bool> indel(sample.size(),false);

      for (int i=0;i<n;i++)
	{
	  if ( badfam.find(sample[i]) != badfam.end() )
	    {
	      indel[i] = true;
	      n_removed++;
	    }
	}

      if (n_removed>0)
	int tmp = deleteIndividuals(indel);
      
      
      printLOG(int2str(f_removed)+" families ( "+
	       int2str(n_removed)+" individuals ) removed due to Mendel errors\n");
      printLOG(int2str(l_removed)+" markers removed due to Mendel errors, "+
	       int2str(nl_all)+" remaining\n");            


      ////////////////////////////
      // Rebuild family structure?
      
      if (f_removed+n_removed>-1)
	{
	  // Wipe existing family structure

	  family.clear();

	  map<string,Individual*> fnd;
	  map<Individual*,int> idmap;

	  linkRelateds(idmap, fnd);
	  parseTrios();      
	}

    }

  
}



void Plink::perm_testTDT(Perm & perm)
{

  //////////////////////////////////
  // Individual-major mode analysis
  
  if (par::SNP_major) SNP2Ind();
  

  /////////////////////
  // Read sets, if any

  if (par::set_test) readSet();
 
  // SET statistics
  Set S(snpset);
  
  // Drop any non-set SNPs, if using 
  // set-based tests
  
  if (par::set_test) 
    S.dropNotSet(*this);
  
  ///////////////////////////////////////////
  // Calculate original results for true data

  vector<bool> dummy(family.size(),false);

  perm.setTests(nl_all);
  perm.setPermClusters(*this);

  vector<double> original = testTDT(true, false, perm, dummy, dummy);

  
  ////////////////////////////  
  // Display corrected p-values?
  
  if (par::multtest)
    {
      vector<double> obp(0);
      for (int l=0; l<nl_all;l++)
	obp.push_back(original[l]);
      multcomp(obp,".tdt");
    }

  ////////////////////////////////
  // If no permutations requested, 
  // we can finish here

  if (!par::permute) return;

  
  //////////////////////
  // Make sets?

  if (par::set_test) S.cumulativeSetSum_WITHLABELS(*this,original);


  //////////////////////
  // Begin permutations
  
  bool finished = false;
  while(!finished)
    {
      
      ///////////////////////////////////
      // Set up permutation list for TDT
      
      // Permutations are constant across family and markers
      // flipA/B[permutation][family]
      
      vector<bool> fA(family.size(),false);
      vector<bool> fP(family.size(),false);
      
      for (int f=0; f<family.size(); f++)
	{
	  if (CRandom::rand() < 0.5) fA[f] = true;
	  if (CRandom::rand() < 0.5) fP[f] = true;
	}
      
      vector<double> pr = testTDT(false, true, perm, fA, fP);
      
      //////////////////////
      // Make sets?

      if (par::set_test)
        S.cumulativeSetSum_WITHOUTLABELS(pr,perm.current_reps()+1);

      ////////////////////////////////
      // Standard permutation counting
      
      finished = perm.update(pr,original);

    } // next permutation

  if (!par::silent)
    cout << "\n\n";


  ///////////////////////////////////////////
  // Calculate SET-based empirical p-values
  
  if (par::set_test) 
    {
      printLOG("Calculating empirical SET-based p-values\n");
      S.empiricalSetPValues();
    }


  ////////////////////
  // Display results
  
  ofstream TDT;
  string f;
  if (par::adaptive_perm) f = par::output_file_name + ".tdt.perm";
  else f = par::output_file_name + ".tdt.mperm";

  TDT.open(f.c_str(),ios::out);
  printLOG("Writing TDT permutation results to [ " + f + " ] \n"); 
  TDT.precision(4);

  TDT << setw(4) << "CHR" << " "
      << setw(par::pp_maxsnp) << "SNP" << " ";
  
  if (par::perm_TDT_basic) TDT << setw(12) << "CHISQ_TDT" << " ";
  else if (par::perm_TDT_parent) TDT << setw(12) << "CHISQ_PAR" << " ";
  else TDT << setw(12) << "CHISQ_COM" << " ";
  
  TDT << setw(12) << "EMP1" << " ";
  if (par::adaptive_perm)
    TDT << setw(12) << "NP" << " " << "\n";  
  else
    TDT << setw(12) << "EMP2" << " " << "\n";  

  for (int l=0; l<nl_all; l++)
    {	
      
      // Skip?, if filtering p-values
      if ( par::pfilter && perm.pvalue(l) > par::pfvalue ) 
	continue;

      TDT << setw(4) << locus[l]->chr << " "
	  << setw(par::pp_maxsnp) << locus[l]->name << " "; 
      if (original[l] < -0.5)
	TDT << setw(12) << "NA"  << " " 
	    << setw(12) << "NA"  << " " 
	    << setw(12) << "NA";
      else
	{
	  TDT << setw(12) << original[l] << " "
	      << setw(12) << perm.pvalue(l) << " ";
	  
	  if (par::adaptive_perm)
	    TDT << setw(12) << perm.reps_done(l);
	  else
	    TDT << setw(12) << perm.max_pvalue(l);
	}
      TDT << "\n";
    }
  
  TDT.close();

  ////////////////////////////  
  // Display SET-based results
 
  if (par::set_test)
    {
      
      f = par::output_file_name + ".tdt.set";
      TDT.open(f.c_str(),ios::out);
      printLOG("Writing set-based TDT results to [ " +f+ " ] \n");
      TDT.clear();

      // Header row
      TDT << setw(12) << "SET" << " "
	  << setw(6) << "S"  << " "
	  << setw(par::pp_maxsnp) << "SNP" << " "
	  << setw(12) << "T" << " "
	  << setw(12) << "P_0" << " "
	  << setw(12) << "P_1" << " "
	  << setw(12) << "P_2" << " "
	  << "\n";

      for (int i=0;i<S.pv_set.size();i++)
	{
	  TDT << "\n";
	  for (int j=0;j<S.pv_set[i].size();j++)
	    {
	      TDT << setw(12) << setname[i]  << " "
		  << setw(6)
		  << string("S<"+int2str(j+1+S.s_min[i])) << " "
		  << setw(par::pp_maxsnp)
		  << S.setsort[i][j]  << " "
		  << setw(12)
		  << S.stat_set[i][j][0] << " "
		  << setw(12)
		  << S.pv_set[i][j][0] << " "
		  << setw(12)
		  << S.pv_maxG_set[i][j]/(par::replicates+1)  << " "
		  << setw(12)
		  << S.pv_maxE_set[i][j]/(par::replicates+1)  << " "
		  << "\n"; 
	    }
	}
      
      TDT.close();
    }


  
}



vector<double> Plink::testTDT(bool print_results, 
			      bool permute,
			      Perm & perm,
			      vector<bool> & flipA,
			      vector<bool> & flipP)
{
  
  
  // TDT and X chromosome: males are coded as homozygous i.e. father
  // should always be uninformative;
  // male child will always receive his X from father
  // females as usual
  
  ///////////////////////////
  // Vector to store results 

  vector<double> res(nl_all);
  double zt;

  ofstream TDT;

  if (print_results)
  {
   string f = par::output_file_name + ".tdt";
   TDT.open(f.c_str(),ios::out);
   printLOG("Writing TDT results (asymptotic) to [ " + f + " ] \n");
   TDT << setw(4) << "CHR" << " " 
       << setw(par::pp_maxsnp) << "SNP" << " "
       << setw(6) << "A1:A2" << " "
       << setw(12) << "T:U_TDT" << " "
       << setw(12) << "OR_TDT" << " ";

   if (par::display_ci)
     TDT << setw(12) << string("L"+dbl2str(par::ci_level*100)) << " "
	 << setw(12) << string("U"+dbl2str(par::ci_level*100)) << " ";

   TDT << setw(12) << "CHISQ_TDT" << " "
       << setw(12) << "P_TDT" << " ";
   
   if (par::discordant_parents)
     TDT << setw(12) << "A:U_PAR" << " "
	 << setw(12) << "CHISQ_PAR" << " "
	 << setw(12) << "P_PAR" << " "
	 << setw(12) << "CHISQ_COM" << " "
	 << setw(12) << "P_COM" << " ";

   TDT << "\n";

   if (par::display_ci)
     zt = ltqnorm( 1 - (1 - par::ci_level) / 2  ) ; 
  }


  ///////////////////////////////////
  // Perform analysis for each locus
  
  for (int l=0; l<nl_all; l++)
    {
                        
      // Adaptive permutation, skip this SNP?
      if (par::adaptive_perm && (!perm.test[l])) 
	continue;

      // Transmission counts
      
      double t1 = 0;
      double t2 = 0;
      
      // Count over families

      for (int f=0; f<family.size(); f++)
	{

	  if ( ! family[f]->TDT ) continue;

	  int trA = 0;  // transmitted allele from first het parent
	  int unA = 0;  // untransmitted allele from first het parent
	  
	  int trB = 0;  // transmitted allele from second het parent
	  int unB = 0;  // untransmitted allele from second het parent
	  
	  Individual * pat = family[f]->pat;
	  Individual * mat = family[f]->mat;
	  vector<Individual *> kid = family[f]->kid;

	  bool pat1 = pat->one[l];
	  bool pat2 = pat->two[l];
 
	  bool mat1 = mat->one[l];
	  bool mat2 = mat->two[l];
	  
	  // We need two genotyped parents, with 
	  // at least one het

	  if ( pat1 == pat2 && 
	       mat1 == mat2 ) 
  	    continue;

	  if ( ( pat1 && !pat2 ) || 
	       ( mat1 && !mat2 ) ) 
	    continue;


	  // Consider all offspring in nuclear family
	  
	  for (int c=0; c<kid.size(); c++)
	    {

	      // Only consider affected children
	      if ( ! kid[c]->aff ) continue;
	      
	      bool kid1 = kid[c]->one[l];
	      bool kid2 = kid[c]->two[l];
	      
	      // Skip if offspring has missing genotype
	      if ( kid1 && !kid2 ) continue;
	      
	      // We've now established: no missing genotypes
	      // and at least one heterozygous parent

	      // Kid is 00

	      if ( (!kid1) && (!kid2) ) 
		{
		  if ( ( (!pat1) && pat2 ) && 
		       ( (!mat1) && mat2 ) )
		    { trA=1; unA=2; trB=1; unB=2; }
		  else 
		    { trA=1; unA=2; } 
		}
	      else if ( (!kid1) && kid2 )  // Kid is 01
		{
		  // het dad
		  if (pat1 != pat2 )
		    {
		      // het mum
		      if ( mat1 != mat2 )
			{ trA=1; trB=2; unA=2; unB=1; }
		      else if ( !mat1 ) 
			{ trA=2; unA=1; }
		      else { trA=1; unA=2; }
		    }
		  else if ( !pat1 ) 
		    {
		      trA=2; unA=1; 
		    }		    
		  else
		    {
		      trA=1; unA=2;
		    }
		}
	      else // kid is 1/1
		{
		  
		  if ( ( (!pat1) && pat2 ) && 
		       ( (!mat1) && mat2 ) )
		    { trA=2; unA=1; trB=2; unB=1; }
		  else 
		    { 
		      trA=2; unA=1;
		    }
		}
	      
	      // We have now populated trA (first transmission) 
	      // and possibly trB also 
	      
	      ////////////////////////////////////////
	      // Permutation? 50:50 flip (precomputed)
	      
	      if (permute) { 
		if (flipA[f]) 
		  { 
		    int t=trA; 
		    trA=unA; 
		    unA=t;  
		
		    t=trB; 
		    trB=unB; 
		    unB=t;  
		  } 
	      }
	      
	      // Increment transmission counts
	      
	      if (trA==1) t1++;
	      if (trB==1) t1++;
	      if (trA==2) t2++;
	      if (trB==2) t2++; 
	      
	    } // next offspring in family
	  
	}  // next nuclear family
      
      

      /////////////////////////////////////////////
      // Consider parental discordance information
      
      double p1 = 0; 
      double p2 = 0;
      double d1 = 0;
      double d2 = 0;

      if (par::discordant_parents)
	{
	  
	  // Count over families
	  
	  for (int f=0; f<family.size(); f++)
	    {

	      // Requires parental discordance...
	      if ( ! family[f]->discordant_parents ) 
		continue;
	      
	      Individual * pat = family[f]->pat;
	      Individual * mat = family[f]->mat;

	      bool pat1 = pat->one[l];
	      bool pat2 = pat->two[l];
	      
	      bool mat1 = mat->one[l];
	      bool mat2 = mat->two[l];
	      
	      
	      // ...and that both are genotyped
	      
	      if ( ( pat1 && !pat2 ) || 
		   ( mat1 && !mat2 ) ) 
		continue;
	      

	      ////////////////////////////////////////
	      // Permutation? 50:50 flip (precomputed)
	      
	      if (permute)
		{
		  if (flipP[f]) 
		    {
		      if (pat->aff) { pat->aff = false; mat->aff = true; }
		      else { pat->aff = true; mat->aff = false; }
		    }
		}
	      
	      // Get number of '0' alleles that the father has above the mother
	      
	      if ( pat1 == mat1 &&
		   pat2 == mat2 ) 
		continue;  // d = 0;
	      else if ( pat->aff ) // affected pat 
		{
		  if ( (!pat1) && (!pat2) ) // F/F
		    {
		      // mat will either be F/T or T/T
		      if ( mat1 ) d1++;
		      else p1++;
		    }
		  else if ( (!pat1 ) && pat2 ) // pat F/T
		    {
		      // mat either F/F or T/T
		      if ( mat1 ) 
			p1++;
		      else
			p2++;	      
		    }
		  else // pat must be T/T
		    {
		      // mat will either be F/F or F/T
		      if ( mat1 ) d2++;
		      else p2++;
		    }
		} 
	      else // affected mat 
		{
		  if ( (!pat1) && (!pat2) ) // F/F
		    {
		      // mat will either be F/T or T/T
		      if ( mat1 ) d2++;
		      else p2++;
		    }
		  else if ( (!pat1 ) && pat2 ) // pat F/T
		    {
		      // mat either F/F or T/T
		      if ( mat1 ) 
			p2++;
		      else
			p1++;	      
		    }
		  else // pat must be T/T
		    {
		      // mat will either be F/F or F/T
		      if ( mat1 ) d1++;
		      else p1++;
		    }
		}
	    }
	  
	}
      
      /////////////////////////////
      // Finished counting: now compute
      // the statistics
      
      double tdt_chisq, par_chisq, com_chisq;
      tdt_chisq = par_chisq = com_chisq = -1;
      
      // Basic TDT test
      
      if (t1+t2 > 0)
	tdt_chisq = ((t1-t2)*(t1-t2))/(t1+t2);
      
      if (par::discordant_parents)
	{
	  // parenTDT 
	  
	  if ( p1+p2+4*(d1+d2) > 0 ) 
	    par_chisq = (((p1+2*d1)-(p2+2*d2))*((p1+2*d1)-(p2+2*d2)))
	      /(p1+p2+4*(d1+d2));
	  
	  
	  // Combined test
	  
	  if ( t1+p1+4*d1+t2+p2+4*d2 > 0 )
	    com_chisq = ( ( (t1+p1+2*d1) - (t2+p2+2*d2) ) 
			  * ( (t1+p1+2*d1) - (t2+p2+2*d2) ) ) 
	      / ( t1+p1+4*d1+t2+p2+4*d2 ) ;
	}
      
      
      // Display asymptotic results
      
      if (print_results)
	{

	  double pvalue = chiprobP(tdt_chisq,1);
	  
	  // Skip?, if filtering p-values
	  if ( par::pfilter && pvalue > par::pfvalue ) 
	    continue;

	  TDT.precision(4);
	  
	  TDT << setw(4) << locus[l]->chr << " " 
	      << setw(par::pp_maxsnp) << locus[l]->name << " " 	
	      << setw(6) 
	      << string(locus[l]->allele1 + ":" + locus[l]->allele2) << " "
	      << setw(12) << dbl2str(t1)+":"+dbl2str(t2) << " ";
     
	  // Odds ratio for T:U
	  double OR = t1 / t2;

	  double zero=0;
	  if (OR == 0 || OR != OR || OR == 1/zero || OR == -1 / zero )
	    {
	      TDT << setw(12) << "NA" << " ";
	      if (par::display_ci) 
		TDT << setw(12) << "NA" << " "
		    << setw(12) << "NA" << " ";
	    }
	  else
	    {
	      TDT << setw(12) << OR << " ";
	      
	      if (par::display_ci)
		{
		  double OR_lower = exp( log(OR) - zt * sqrt(1/t1+1/t2)) ;
		  double OR_upper = exp( log(OR) + zt * sqrt(1/t1+1/t2)) ;
		  
		  TDT << setw(12) << OR_lower << " "
		      << setw(12) << OR_upper << " ";
		}
	    }

	  if (tdt_chisq>=0)
	    TDT << setw(12) << tdt_chisq << " "
		<< setw(12) << chiprobP(tdt_chisq,1) << " ";
	  else
	    TDT << setw(12) << "NA" << " "
		<< setw(12) << "NA" << " "; 
	  
	  if (par::discordant_parents)
	    {
	      TDT << setw(12) 
		  <<  dbl2str(p1+2*d1)+":"+dbl2str(p2+2*d2) << " ";
	      
	      if (par_chisq>=0)
		TDT << setw(12) << par_chisq  << " "
		    << setw(12) << chiprobP(par_chisq,1) << " ";
	      else
		TDT << setw(12) << "NA" << " "	
		    << setw(12) << "NA" << " "; 
	      
	      if (com_chisq>=0)
		TDT << setw(12) << com_chisq << " "
		    << setw(12) << chiprobP(com_chisq,1) << " ";
	      else
		TDT << setw(12) << "NA" << " "
		    << setw(12) << "NA" << " "; 
	    }
	  TDT << "\n";
	}


      ///////////////////////////////////////////
      // Choose which statistic for permutation

      if (par::perm_TDT_basic) res[l] = tdt_chisq;
      else if (par::perm_TDT_parent) res[l] = par_chisq;
      else res[l] = com_chisq;
    
    } // next locus


  //////////////////////////////
  // Close output file, if open

  if (print_results)
    TDT.close();

  ///////////////////////////////////////////
  // Return chosen statistic for permutation

  return res;

}
